/*
 LiPo Battery Storage Monitor - Sensor
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 19/11/2017
 Last Modified: 13/03/2018
 
 Read the LiPo cell voltages, transmit the voltages to the server and then sleep for a certain number of hours.
 The sensor number, link byte and delay hours is programmed by holding the link button at startup 
 and accessing the server's /admin page
 
 Program the ATtiny24/44/84 with internal oscillator startup time of 4ms
 avrdude -p ATtiny84 -c usbasp -U lfuse:w:0x52:m -U hfuse:w:0xdf:m -U efuse:w:0xff:m 
 
 */
 
// ATtiny24/44/84 Pin map
//
//                                +-\/-+
//                          VCC  1|o   |14  GND
//               nRF24 MISO PB0  2|    |13  PA0 1S / Button with 10K resistor
//               nRF24 MOSI PB1  3|    |12  PA1 2S
//                    Reset PB3  4|    |11  PA2 3S
//                nRF24 SCK PB2  5|    |10  PA3 4S
//         	      nRF24 CSN PA7  6|    |9   PA4 5S
//   nRF24 CE / N Fet / LED PA6  7|    |8   PA5 6S 
//                                +----+

#define F_CPU 1000000 // 1 MHz

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include "setup.c"
#include "mirf.c"

int main(void) {
	setup();
	check_link_button();
	
	while(1) {
		// Check battery voltage
		mirf_CE_hi; // N mosfets on to read battery voltages
		sbi(ADCSRA, ADEN); // Turn on ADC
		adc_read(0); // First reading is ignored
		
		uint8_t count = 2;
		for (uint8_t x = 0; x < 6; x++) {
			int batteryVoltage = adc_read(x);
			data_in_out[count++] = batteryVoltage >> 8;
			data_in_out[count++] = batteryVoltage & 0xFF;
		}
		cbi(ADCSRA, ADEN); // Turn off ADC
		mirf_CE_lo;
		
		// Include the sensor number and link byte
		data_in_out[SENSOR_NUMBER] = sensorNumber;
		data_in_out[LINK_BYTE] = linkByte;
		
		// Transmit and power down afterwards
		mirf_transmit_data();
		
		// Wait the delay hours
		uint16_t secondsCount = 0;
		uint8_t hoursCount = 0;
		while (hoursCount < hoursDelay) {
			watchdog_sleep(T8S);
			secondsCount += 8;
			if (secondsCount >= 3600) {
				secondsCount = 0;
				hoursCount += 1;
			}
		}
	}
}

EMPTY_INTERRUPT(WDT_vect);
EMPTY_INTERRUPT(ADC_vect);
