/*
 LiPo Battery Storage Monitor - Sensor
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 19/11/2017
 Last Modified: 13/03/2018
 
 */
 
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

// Watchdog timeouts to values
#define T16MS 0
#define T32MS 1
#define T64MS 2
#define T128MS 3
#define T250MS 4
#define T500MS 5
#define T1S 6
#define T2S 7
#define T4S 8
#define T8S 9

// nRF24 pin def
#define CE  PA6
#define CSN PA7
#define SCK PB2
#define MOSI PB1 // ATtiny MOSI
#define MISO PB0 // ATtiny MISO

// Data in/out elements
#define SENSOR_NUMBER 0
#define LINK_BYTE 1
#define HOURS_DELAY 2

// EEPROM variable locations
#define sensorNumberLocation 1
#define linkByteLocation 2
#define hoursDelayLocation 3

uint8_t data_in_out[14];
uint8_t sensorNumber = 25;
uint8_t linkByte = 0;
uint8_t hoursDelay = 1;

extern void mirf_init(void);
extern void mirf_config(void);
extern uint8_t mirf_receive_data(void);

// Bit-bang SPI
uint8_t spi_transfer(uint8_t data) {
	uint8_t dataRead = 0;
	for (int8_t i = 7; i >= 0; i--) {
		if (data & (1<<i)) { // Output
			PORTB |= (1<<MOSI);
		} 
		else {
			PORTB &= ~(1<<MOSI);
		}
		
		PORTB |= (1<<SCK);
		
		if (PINB & (1<<MISO)) { // Input
			dataRead |= (1<<i);
		}
		PORTB &= ~(1<<SCK);
	}
	
	return dataRead;
}

// Write data using SPI
void spi_write_data(uint8_t * dataout, uint8_t len) {
	for (uint8_t i = 0; i < len; i++) {
		spi_transfer(dataout[i]);
	}
}

// Read data using SPI
void spi_read_data(uint8_t * datain, uint8_t len) {
	for (uint8_t i = 0; i < len; i++) {
		datain[i] = spi_transfer(0x00);
	}
}

// Sleep
void system_sleep(void) {
	set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
	sleep_mode(); // System sleeps here
}

// Setup watchdog timed sequence
void setup_watchdog(uint8_t wdDelay) {
	if (wdDelay > 7) {
		wdDelay = wdDelay & 7;
		wdDelay |= (1<<5); // Set WDP3 bit
	}
	
	MCUSR &= ~(1<<WDRF); // Turn off flag
	WDTCSR |= (1<<WDCE) | (1<<WDE); // Start timed sequence
	WDTCSR = wdDelay | (1<<WDIE); // Set new watchdog timeout value
}

// Turn off the Watchdog
void turn_off_watchdog(void) {
	cli(); // Turn interrupts off
	MCUSR &= ~(1<<WDRF);
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	WDTCSR = 0;
	sei(); // Turn interrupts back on
}

// Watchdog sleep function - setup, sleep and turn off
void watchdog_sleep(uint8_t timeout) { 
	setup_watchdog(timeout);
	system_sleep();
	turn_off_watchdog();
}

// ADC Read
int adc_read(uint8_t pin) {
	uint8_t low, high;
	
	ADMUX = (1 << 7) | (pin & 0x3f); // 1.1V voltage reference
	
	set_sleep_mode(SLEEP_MODE_ADC); // Set sleep mode ADC
	sleep_mode(); // System sleeps here
	
	// Read ADCL first
	low = ADCL;
	high = ADCH;
	
	// Combine the two bytes
	return (high << 8) | low;
}

// Check if the button is pressed (via the ADC), if so, stay in listen mode until we hear from the server 
// what our sensor number is, link byte and delay hours to use
void check_link_button(void) {
	sbi(ADCSRA, ADEN); // Turn on ADC
	adc_read(0); // First reading is ignored
	uint16_t buttonAdc = adc_read(0);
	cbi(ADCSRA, ADEN); // Turn off ADC
	
	// Button would read 812mV in the worst case
	if (buttonAdc >= 512) {
		mirf_receive_data();
		
		// Update variables/EEPROM
		sensorNumber = data_in_out[SENSOR_NUMBER];
		linkByte = data_in_out[LINK_BYTE];
		hoursDelay = data_in_out[HOURS_DELAY];
		
		eeprom_write_byte((uint8_t*) sensorNumberLocation, sensorNumber);
		eeprom_write_byte((uint8_t*) linkByteLocation, linkByte);
		eeprom_write_byte((uint8_t*) hoursDelayLocation, hoursDelay);
	}
}

// Setup
void setup(void) {
	_delay_ms(250); // Wait a little bit after power on
	
	// mIRF initialise and configure
	mirf_init();
	mirf_config();
	
	// Set ADC prescale factor to 8
	// 1 MHz / 8 = 125 KHz, inside the desired 50-200 KHz range.
	sbi(ADCSRA, ADPS1);
	sbi(ADCSRA, ADPS0);
	
	// Turn on ADC interrupt
	sbi(ADCSRA, ADIE);
	
	// Load sensor number, link byte and delay hours from EEPROM if programmed
	uint8_t checkEepromByte = eeprom_read_byte((uint8_t*) sensorNumberLocation);
	if (checkEepromByte != 255) {
		sensorNumber = eeprom_read_byte((uint8_t*) sensorNumberLocation);
		linkByte = eeprom_read_byte((uint8_t*) linkByteLocation);
		hoursDelay = eeprom_read_byte((uint8_t*) hoursDelayLocation);
	}
	
	sei(); // Turn on interrupts
}
