/*
 LiPo Battery Storage Monitor - Server
 Version: 1.1
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 19/11/2017
 Last Modified: 13/03/2018
 */
 
#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include "esp8266-sendemail/sendemail.h"

#include <SPI.h>
#include <EEPROM.h>
#include "mirf.h"
#include "nRF24L01.h"

ESP8266WebServer server ( 80 );

// EEPROM locations
#define EEPROM_LOCATION_WIFI_SSID 0
#define EEPROM_LOCATION_WIFI_PASSWORD 50
#define EEPROM_LOCATION_EMAIL_ENABLED 148
#define EEPROM_LOCATION_EMAIL_PORT 149
#define EEPROM_LOCATION_EMAIL_SERVER 150
#define EEPROM_LOCATION_EMAIL_FROM 200
#define EEPROM_LOCATION_EMAIL_FROM_NAME 250
#define EEPROM_LOCATION_EMAIL_TO 300
#define EEPROM_LOCATION_VOLT_MIN 400
#define EEPROM_LOCATION_VOLT_MAX 410
#define EEPROM_LOCATION_DEVICE_SETUP 420

// Vars for EEPROM
char wifiSsid[50];
char wifiPassword[80];
uint8_t emailEnabled = 0;
uint8_t emailPort = 0;
char emailServer[50];
char emailFrom[50];
char emailFromName[50];
char emailTo[50];
double voltageMin = 3.7;
double voltageMax = 4.35;

// Vars misc
const int led = 2;
const int ap_mode_pin = 5;
uint32_t currentUptime = 0;
String tempCalcTimeStr = "";
uint8_t data_in_out[14];
uint8_t linkByte = 49;
char nrfLog[50][30];
uint8_t nrfLogCounter = 0;

// Sensors structure
#define MAX_SENSORS 30
struct Sensor {
  uint8_t emailSent;
  uint32_t emailSentTime;
  uint32_t lastseen;
};
Sensor sensorList[MAX_SENSORS];

// Voltage structure
double sensorVoltage[MAX_SENSORS][6];


// Increment uptime every second
void timer0_ISR (void) {
  timer0_write(ESP.getCycleCount() + 80000000L); // 80MHz = 1 sec
  currentUptime++;
}

// Calculate the last check in time
void calculate_last_time(uint32_t sensorTime) {  
  uint32_t calcTime = currentUptime - sensorTime;
  uint8_t calcDays = 0;
  uint8_t calcHours = 0;
  uint8_t calcMinutes = 0;
  uint8_t calcSeconds = 0;
  
  tempCalcTimeStr = "";

  if (sensorTime >= 1) {
    if (calcTime / 86400 >= 1) {
      calcDays = calcTime / 86400;
      calcTime = calcTime - (86400 * calcDays);
      tempCalcTimeStr += calcDays;
      tempCalcTimeStr += "d, ";
    }
    if (calcTime / 3600 >= 1) {
      calcHours = calcTime / 3600;
      calcTime = calcTime - (3600 * calcHours);
      tempCalcTimeStr += calcHours;
      tempCalcTimeStr += "h, ";
    }
    if (calcTime / 60 >= 1) {
      calcMinutes = calcTime / 60;
      calcTime = calcTime - (60 * calcMinutes);
      tempCalcTimeStr += calcMinutes;
      tempCalcTimeStr += "m, ";
    }
    calcSeconds = calcTime;
    
    tempCalcTimeStr += calcSeconds;
    tempCalcTimeStr += "s";
  }
}


// When ESP8266 is in setup mode, it's an AP, allow us to enter in Wifi details
void handleAP() {
  digitalWrite ( led, 1 );

  String temp = "<html>\
<head>\
<title>LiPo Battery Monitor</title>\
</head>\
<link rel='stylesheet' type='text/css' href='style.css'>\
<body>";
   
  temp += "<form method='POST' action=''>\
  <p align='center'><font class='normaltext-css'><b>Connect to your Wifi Access Point</b><br>\
 Please enter in your Wifi SSID and password then press Submit.</font></p>\
  <p align='center'><font class='normaltext-css'>SSID <input type='text' name='ssid' size='20'></font></p>\
  <p align='center'><font class='normaltext-css'>Password <input type='text' name='password' size='20'></font></p>\
  <p align='center'><font class='normaltext-css'><input type='submit' value='Submit' name='submit'></font></p>\
  </form><br/>";
  
  if (server.hasArg("ssid")) {
    strncpy(wifiSsid, server.arg("ssid").c_str(), 49);
    strncpy(wifiPassword, server.arg("password").c_str(), 79);

    temp += "<p align='center'><font class='normaltext-css'>Completed. Please reboot the device.<br/>Access the device after the reboot here: <a href='http://LiPoBatMon.local'>http://LiPoBatMon.local</a></font></p>";

    // Save to EEPROM
    uint8_t eepromCounter = 0;
    for (uint8_t x = 0; x < 50; x++) {
      EEPROM.write(eepromCounter++, wifiSsid[x]);
    }
    eepromCounter++;
    for (uint8_t x = 0; x < 80; x++) {
      EEPROM.write(eepromCounter++, wifiPassword[x]);
    }

    // Load some default values in EEPROM
    EEPROM.write(EEPROM_LOCATION_EMAIL_ENABLED, 0);
    EEPROM.write(EEPROM_LOCATION_EMAIL_PORT, 25);
    EEPROM.write(EEPROM_LOCATION_EMAIL_SERVER, 0);
    EEPROM.write(EEPROM_LOCATION_EMAIL_FROM, 0);
    EEPROM.write(EEPROM_LOCATION_EMAIL_FROM_NAME, 0);
    EEPROM.write(EEPROM_LOCATION_EMAIL_TO, 0);
    EEPROM.put(EEPROM_LOCATION_VOLT_MIN, voltageMin);
    EEPROM.put(EEPROM_LOCATION_VOLT_MAX, voltageMax);

    // Device setup
    EEPROM.write(EEPROM_LOCATION_DEVICE_SETUP, 1);
    
    EEPROM.commit();
  }
  
  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}


void handleRoot() {
  digitalWrite ( led, 1 );
  
  String temp = "<html><head><title>LiPo Monitor</title></head>\
  <link rel='stylesheet' type='text/css' href='style.css'><body>\
  <div class='text-css'><p align='center'><b>LiPo Monitor</b></p></div>\
  <div align='center'>";

  // POST: Clear sensors or alerts
  if (server.hasArg("clear-sensors")) {
    for (uint8_t x = 0; x < MAX_SENSORS; x++) {
      sensorList[x].lastseen = 0;
    }
    temp += "Sensor list cleared.<br/>";
  }
  else if (server.hasArg("clear-alerts")) {
    for (uint8_t x = 0; x < MAX_SENSORS; x++) {
      sensorList[x].emailSent = 0;
    }
    temp += "Alerts cleared<br/>";
  }
  
  temp += "<table border='1' width='75%' cellspacing='0' cellpadding='0' bordercolorlight='#000000'>\
    <tr>\
      <td align='center' width='7%'><font class='normaltext-css'><b>No</b></font></td>\
      <td align='center'><font class='normaltext-css'><b>Cells</b></font></td>\
      <td align='center' width='12%'><font class='normaltext-css'><b>Status</b></font></td>\
      <td align='center' width='50%'><font class='normaltext-css'><b>Voltage</b></font></td>\
      <td align='center' width='21%'><font class='normaltext-css'><b>Last check in</b></font></td>\
    </tr>";

  for (uint8_t x = 0; x < MAX_SENSORS; x++) {
    if (sensorList[x].lastseen > 0) {

      uint8_t cells = 0;
      uint8_t cellsOk = 1;

      // Check for cell count
      for (uint8_t v = 0; v < 6; v++) {
        if (sensorVoltage[x][v] >= 1) {
          cells = v+1;
        }
      }

      // Check voltage on cells
      for (uint8_t v = 0; v < cells; v++) {
        if (sensorVoltage[x][v] <= voltageMin || sensorVoltage[x][v] >= voltageMax) {
           cellsOk = 0;
        }
      }

      calculate_last_time(sensorList[x].lastseen);

      temp += "<tr>\
      <td align='center' width='13%' height='35'><font class='normaltext-css'>";
      temp += x;
      temp += "</font></td><td align='center' height='35'><font class='normaltext-css'>";
      temp += cells;
      temp += "</font></td><td align='center' height='35'>\
      <font class='normaltext-css'>";
      
      if (cellsOk == 0) {
        temp += "Warning";
      }
      else {
        temp += "OK";
      }
      temp += "</font></td>\
      <td align='center' height='35'><font class='normaltext-css'>";
      
      for (uint8_t v = 0; v < cells; v++) {
        temp += sensorVoltage[x][v];
        temp += "V";
        
        if (v != (cells-1)) {
          temp += ", ";
        }
      }

      temp += "</font></td>\
      <td align='center' height='35'><font class='normaltext-css'>";
      temp += tempCalcTimeStr;
      temp += "</font></td></tr>";
    }
  }

  temp += "</table></div><div align='center'><br/><br/>";
  temp += "<form method='POST' action=''><input type='submit' name='clear-alerts' value='Clear alerts' /></form>";
  temp += "<br/><form method='POST' action=''><input type='submit' name='clear-sensors' value='Clear list' /></form>";
  temp += "</div></body></html>";
  
  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}

void handleLog() {
  digitalWrite ( led, 1 );

  String temp = "<html>\
  <head>\
  <title>LiPo Monitor</title>\
  </head>\
  <link rel='stylesheet' type='text/css' href='style.css'>\
  <body><p align='center'><b>Log</b><br/><br/>";

  // POST: Clear
  if (server.hasArg("clear")) {
    nrfLogCounter = 0;
    for (uint8_t x = 0; x < 50; x++) {
      nrfLog[x][0] = 0;
    }
    temp += "Cleared.<br/>";
  }

  for (uint8_t x = 0; x < 50; x++) {
    if (nrfLog[x][0] >= 1) {
      temp += nrfLog[x];
      temp += "<br/>";
    }
  }

  temp += "<br/><form method='POST' action=''><p align='center'><input type='submit' name='clear' value='Clear' /></form>";
  temp += "</div></body></html>";

  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}


void handleCss() {
  digitalWrite ( led, 1 );
  
  String temp(F(".arm-css {\
  display: inline-block;\
  -webkit-box-sizing: content-box;\
  -moz-box-sizing: content-box;\
  box-sizing: content-box;\
  width: 16%;\
  height: 5%;\
  cursor: pointer;\
  padding: 10px 20px;\
  border: 1px solid #000000;\
  -webkit-border-radius: 3px;\
  border-radius: 3px;\
  font: normal 30px/normal 'Trebuchet MS', Helvetica, sans-serif;\
  color: rgba(255,255,255,0.9);\
  -o-text-overflow: clip;\
  text-overflow: clip;\
  background: rgba(232,126,39,0.66);\
  -webkit-box-shadow: 2px 2px 2px 0 rgba(0,0,0,0.2) ;\
  box-shadow: 2px 2px 2px 0 rgba(0,0,0,0.2) ;\
  -webkit-transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
  -moz-transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
  -o-transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
  transition: all 300ms cubic-bezier(0.42, 0, 0.58, 1);\
}\
.text-css {\
  -webkit-box-sizing: content-box;\
  -moz-box-sizing: content-box;\
  box-sizing: content-box;\
  border: none;\
  font: normal 40px/1 'Trebuchet MS', Helvetica, sans-serif;\
  color: black;\
  -o-text-overflow: ellipsis;\
  text-overflow: ellipsis;\
}\
.normaltext-css {\
  -webkit-box-sizing: content-box;\
  -moz-box-sizing: content-box;\
  box-sizing: content-box;\
  border: none;\
  font: normal 30px/1 'Trebuchet MS', Helvetica, sans-serif;\
  color: black;\
  -o-text-overflow: ellipsis;\
  text-overflow: ellipsis;\
}"));

  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}


void handleAdmin() {
  digitalWrite ( led, 1 );

  String temp = "<html><head><title>LiPo Battery Monitor - Admin</title></head>\
  <link rel='stylesheet' type='text/css' href='style.css'>\
  <body>";

  // Sensor linking text
  temp += "<form method='POST' action=''>\
  <p align='center'><b>Sensor Linking<br>\
  </b>Hold the link button on the sensor and then connect the power.<br>\
  Enter the sensor number you wish to assign it below and press Submit.<br>\
  <p align='center'>Sensor Number <input type='text' name='sensorno' size='4'><br/>\
  Check in delay (1 to 255 hours, 20% -/+) <input type='text' name='hours' size='4'><br/><br/>\
  <input type='submit' value='Submit' name='submit'>\
  </form><br/>";


  // POST: Sensor linking
  if (server.hasArg("sensorno")) {
    uint8_t sensorNo = server.arg("sensorno").toInt();
    uint8_t hoursDelay = server.arg("hours").toInt();
    
    data_in_out[SENSOR_NUMBER] = sensorNo;
    data_in_out[LINK_BYTE] = linkByte;
    data_in_out[HOURS_DELAY] = hoursDelay;

    // Send it a few times to make sure it's received
    for (uint8_t x = 0; x < 10; x++) {
      mirf_CE_lo; // Stop listening
      
      // Transmit the sensor number and link byte to the sensor
      NRF_TX_POWERUP; // Power up to transmitter mode
      delay(3);
      mirf_flush_rx_tx(); // Flush RX and TX FIFO
      
      mirf_CSN_lo;  // Pull down chip select
      SPI.transfer(W_TX_PAYLOAD); // Write cmd to write payload
      spi_write_data(data_in_out, mirf_PAYLOAD); // Write payload
      mirf_CSN_hi; // Pull up chip select
      
      mirf_CE_hi; // Start transmission
      delay(1);
      mirf_CE_lo;
      delay(1); // Wait for transmission
      
      while (!(mirf_status() & (1<<TX_DS))) { // Wait until packet is sent
        if (mirf_status() & (1<<MAX_RT)) { // If maximum retries reached
          mirf_config_register(STATUS, 1<<MAX_RT); // Reset maximum retries to be able to transmit again next time
        }
        delay(1);
      }
      mirf_config_register(STATUS,(1<<TX_DS)); // Reset status register

      delay(50);
    }
    
    // Go back to receiver mode
    NRF_RX_POWERUP; // Power up to receiver mode
    delay(3);
    mirf_flush_rx_tx(); // Flush RX and TX FIFO
    mirf_CE_hi; // Start listening
    
    temp += "<p align='center'>Sensor number programmed: ";
    temp += sensorNo;
    temp += "<br/>Check in delay programmed: ";
    temp += hoursDelay;
    temp += "<br/></p>";
  }


  // POST: Voltage min/max text
  if (server.hasArg("voltage-min")) {
    voltageMin = server.arg("voltage-min").toFloat();
    voltageMax = server.arg("voltage-max").toFloat();

    temp += "<p align='center'><b>Voltage settings changed.</b><br/>";
    temp += "Voltage Min:";
    temp += voltageMin;
    temp += "<br/>Voltage Max:";
    temp += voltageMax;
    temp += "<br/></p>";
    
    EEPROM.put(EEPROM_LOCATION_VOLT_MIN, voltageMin);
    EEPROM.put(EEPROM_LOCATION_VOLT_MAX, voltageMax);
    EEPROM.commit();
  }
  
  temp += "<form method='POST' action=''>\
  <p align='center'><b>Voltage Min / Max</b><br>\
  The minimum and maximum voltage of any cell before receiving a warning/alert.<br/>\
  Voltage Minimum <input type='text' name='voltage-min' size='4' value='";
  temp += voltageMin;
  temp += "'>V<br/>Voltage Maximum <input type='text' name='voltage-max' size='4' value='";
  temp += voltageMax;
  temp += "'>V<br/><br/>\
  <input type='submit' value='Submit' name='submit'></form><br/>";


  // POST: Email alerts
  if (server.hasArg("email-server")) {
    emailEnabled = server.arg("email-enabled").toInt();
    emailPort = server.arg("email-port").toInt();
    strncpy(emailServer, server.arg("email-server").c_str(), 49);
    strncpy(emailFrom, server.arg("email-from").c_str(), 49);
    strncpy(emailFromName, server.arg("email-name").c_str(), 49);
    strncpy(emailTo, server.arg("email-to").c_str(), 49);
    
    temp += "<p align='center'><b>Email settings changed.</b><br/>";
    temp += "Email enabled:";
    temp += emailEnabled;
    temp += "Email server:";
    temp += emailServer;
    temp += "<br/>Email port:";
    temp += emailPort;
    temp += "<br/>Email from:";
    temp += emailFrom;
    temp += "<br/>Email from name:";
    temp += emailFromName;
    temp += "<br/>Email to:";
    temp += emailTo;
    temp += "<br/></p>";
    
    // Save to EEPROM
    int eepromCounter = EEPROM_LOCATION_EMAIL_ENABLED;
    EEPROM.write(eepromCounter++, emailEnabled);
    EEPROM.write(eepromCounter++, emailPort);
    for (uint8_t x = 0; x < 50; x++) {
      EEPROM.write(eepromCounter++, emailServer[x]);
    }
    eepromCounter = EEPROM_LOCATION_EMAIL_FROM;
    for (uint8_t x = 0; x < 50; x++) {
      EEPROM.write(eepromCounter++, emailFrom[x]);
    }
    eepromCounter = EEPROM_LOCATION_EMAIL_FROM_NAME;
    for (uint8_t x = 0; x < 50; x++) {
      EEPROM.write(eepromCounter++, emailFromName[x]);
    }
    eepromCounter = EEPROM_LOCATION_EMAIL_TO;
    for (uint8_t x = 0; x < 50; x++) {
      EEPROM.write(eepromCounter++, emailTo[x]);
    }
  
    EEPROM.commit();
  }

  
  temp += "<form method='POST' action=''>\
  <p align='center'><b>Email alerts on voltage lower than ";
  temp += voltageMin;
  temp += "V or higher than ";
  temp += voltageMax;
  temp += "V</b><br/>";
  temp += "Email enabled: <input type='checkbox' name='email-enabled' value='1'";

  if (emailEnabled == 1) {
    temp += " checked";
  }
  temp += "><br/>\
  Email server: <input type='text' name='email-server' size='10' value='";
  temp += emailServer;
  temp += "'><br/>\
  Email port: <input type='text' name='email-port' size='5' value='";
  temp += emailPort;
  temp += "'><br/>\
  Email from: <input type='text' name='email-from' size='10' value='";
  temp += emailFrom;
  temp += "'><br/>\
  Email from name: <input type='text' name='email-name' size='10' value='";
  temp += emailFromName;
  temp += "'><br/>\
  Email to: <input type='text' name='email-to' size='10' value='";
  temp += emailTo;
  temp += "'><br/>\
  <br/><br/>\
  <input type='submit' value='Submit' name='submit'>\
  </form><br/>";

  temp += "</p></body></html>";
  server.send ( 200, "text/html", temp );
  digitalWrite ( led, 0 );
}

void handleNotFound() {
  digitalWrite ( led, 1 );
  
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += ( server.method() == HTTP_GET ) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";

  for ( uint8_t i = 0; i < server.args(); i++ ) {
    message += " " + server.argName ( i ) + ": " + server.arg ( i ) + "\n";
  }

  server.send ( 404, "text/plain", message );
  digitalWrite ( led, 0 );
}


void setup ( void ) {
  pinMode(led, OUTPUT);
  digitalWrite(led, 0);
  Serial.begin(115200);
  EEPROM.begin(4096);

  // If button is held when booting or if setup byte in EEPROM is 255, boot AP mode
  uint8_t deviceSetup = EEPROM.read(EEPROM_LOCATION_DEVICE_SETUP);
  if (digitalRead(ap_mode_pin) == 1 || deviceSetup != 1) {
    Serial.println("\nAP Mode");
    
    IPAddress Ip(192, 168, 0, 1);
    IPAddress NMask(255, 255, 255, 0);
    WiFi.softAPConfig(Ip, Ip, NMask);
    
    WiFi.mode(WIFI_AP);
    WiFi.hostname("LiPoBatMon");
    WiFi.softAP("LiPoBatMon-AP", "1928472742"); // Change this for every device

    server.on ( "/", handleAP );
    server.on ( "/style.css", handleCss );
    server.onNotFound ( handleNotFound );
    server.begin();

    mirf_setup();
    
    Serial.println("\nReady");
  }
  else {
    Serial.println("\nStation Mode");

    // Load Wifi SSID and password from EEPROM
    int eepromCounter = 0;
    for (uint8_t x = 0; x < 50; x++) {
      wifiSsid[x] = EEPROM.read(eepromCounter++);
    }
    eepromCounter++;
    for (uint8_t x = 0; x < 80; x++) {
      wifiPassword[x] = EEPROM.read(eepromCounter++);
    }

    // Load settings
    eepromCounter = EEPROM_LOCATION_EMAIL_ENABLED;
    emailEnabled = EEPROM.read(eepromCounter++);
    emailPort = EEPROM.read(eepromCounter++);
    for (uint8_t x = 0; x < 50; x++) {
      emailServer[x] = EEPROM.read(eepromCounter++);
    }
    eepromCounter = EEPROM_LOCATION_EMAIL_FROM;
    for (uint8_t x = 0; x < 50; x++) {
      emailFrom[x] = EEPROM.read(eepromCounter++);
    }
    eepromCounter = EEPROM_LOCATION_EMAIL_FROM_NAME;
    for (uint8_t x = 0; x < 50; x++) {
      emailFromName[x] = EEPROM.read(eepromCounter++);
    }
    eepromCounter = EEPROM_LOCATION_EMAIL_TO;
    for (uint8_t x = 0; x < 50; x++) {
      emailTo[x] = EEPROM.read(eepromCounter++);
    }
    
    EEPROM.get(EEPROM_LOCATION_VOLT_MIN, voltageMin);
    EEPROM.get(EEPROM_LOCATION_VOLT_MAX, voltageMax);

    //Serial.print("SSID: ");
    //Serial.println(wifiSsid);
    //Serial.print("Password: ");
    //Serial.println(wifiPassword);

    Serial.print("Email Enabled: ");
    Serial.println(emailEnabled);
    Serial.print("Email server: ");
    Serial.println(emailServer);
    Serial.print("Email Port: ");
    Serial.println(emailPort);
    Serial.print("Email from: ");
    Serial.println(emailFrom);
    Serial.print("Email from name: ");
    Serial.println(emailFromName);
    Serial.print("Email to: ");
    Serial.println(emailTo);

    // Connect
    WiFi.mode(WIFI_STA);
    WiFi.hostname("LiPoBatMon");
    WiFi.begin (wifiSsid, wifiPassword);
    
    // Wait for connection
    Serial.print("Connecting to Wifi");
    while (WiFi.status() != WL_CONNECTED) {
      delay(500);
      Serial.print(".");
    }
    Serial.println("\nConnected");
    Serial.print("IP address:\t");
    Serial.println(WiFi.localIP()); 
    MDNS.begin("LiPoBatMon");
  
    server.on ( "/", handleRoot );
    server.on ( "/admin", handleAdmin );
    server.on ( "/style.css", handleCss );
    server.onNotFound ( handleNotFound );
    server.begin();
    
    mirf_setup();
    
    Serial.println("\nReady");
  
    noInterrupts();
    timer0_isr_init();
    timer0_attachInterrupt(timer0_ISR);
    timer0_write(ESP.getCycleCount() + 80000000L); // 80MHz = 1 sec
    interrupts();
  }
}

void loop ( void ) {
  server.handleClient();

  // Check if we have a packet waiting
  if (mirf_status() & (1<<RX_DR)) { // Wait until data is ready to be read
    mirf_CSN_lo; // Pull down chip select
    SPI.transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
    spi_read_data(data_in_out, mirf_PAYLOAD); // Read payload
    mirf_CSN_hi; // Pull up chip select
    mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register

    mirf_flush_rx_tx(); // Flush RX and TX FIFO

    uint8_t sensorNo = data_in_out[SENSOR_NUMBER];
    for (uint8_t x = 0; x < 14; x++) {
      Serial.print(data_in_out[x]);
      Serial.print(" ");
    }

    // Back to counter 0 and clear
    if (nrfLogCounter > 48) {
      nrfLogCounter = 0;
      for (uint8_t x = 0; x < 50; x++) {
        nrfLog[x][0] = 0;
      }
    }

    sprintf (nrfLog[nrfLogCounter++], "Sensor #%X, Link byte: %X", data_in_out[SENSOR_NUMBER], data_in_out[LINK_BYTE]);
   
    // If link byte matches
    if (data_in_out[LINK_BYTE] == linkByte) {
      sensorList[sensorNo].lastseen = currentUptime;
  
      uint8_t arrayCounter = 2;
      for (uint8_t x = 0; x < 6; x++) {
        double voltage = ((data_in_out[arrayCounter] << 8) | data_in_out[arrayCounter+1]) * 0.00107421875 * 26.6382;
        arrayCounter += 2;
        sensorVoltage[sensorNo][(x)] = voltage;
      }
      
      // Check for cell count
      uint8_t cells = 0;
      for (uint8_t v = 0; v < 6; v++) {
        if (sensorVoltage[sensorNo][v] >= 1) {
          cells = v+1;
        }
      }

      // Calculate voltage on cells
      for (uint8_t v = 0; v < cells; v++) {
        if (v >= 1) {
          for (int8_t c = v-1; c >= 0; c--) {
            sensorVoltage[sensorNo][v] -= sensorVoltage[sensorNo][c];
          }
        }
      }

      // Clear alerts automatically after 24 hours of a sent email
      if (emailEnabled == 1 && sensorList[sensorNo].emailSent == 1) {
        if ((currentUptime - sensorList[sensorNo].emailSentTime) > 86400) {
          sensorList[sensorNo].emailSent = 0;
        }
      }
      
      // If email alert enabled, check voltage on cells
      if (emailEnabled == 1 && sensorList[sensorNo].emailSent == 0) {
        for (uint8_t v = 0; v < cells; v++) {
          if (sensorVoltage[sensorNo][v] <= voltageMin || sensorVoltage[sensorNo][v] >= voltageMax) {         
            char emailSubject[30];
            char emailText[50];
            char cellVoltage[10];
            dtostrf(sensorVoltage[sensorNo][v], 4, 2, cellVoltage);

            sprintf(emailSubject, "LiPo Warning - Bat #%i", sensorNo);
            sprintf(emailText, "Cell %i at %sV", v+1, cellVoltage);
            
            SendEmail e(emailServer, emailPort, "", "", 6000, false);
            e.send(emailFromName, emailFrom, emailTo, emailSubject, emailText);
            sensorList[sensorNo].emailSent = 1;
            sensorList[sensorNo].emailSentTime = currentUptime;

            Serial.print("Email subject");
            Serial.println(emailSubject);
            Serial.print("Email Text");
            Serial.println(emailText);
            Serial.println("Email sent");
            
            break; // Don't send any more for this battery
          }
        }
      }
    }
    
    Serial.print("Uptime ");
    Serial.print(currentUptime);
    Serial.println("");
    delay(50);
  }
}


